--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        toolset.lua
--

-- imports
import(".api_checker")

function main(opt)
    opt = opt or {}
    -- we only support one tool for each tool kind, e.g. set_toolset("cxx", "clang")
    -- @see https://github.com/xmake-io/xmake/pull/7597
    api_checker.check_targets("toolset", table.join(opt, {keyvalues = true, check = function(target, toolkind, program)
        if type(program) ~= "string" then
            return false, string.format("set_toolset(\"%s\", \"%s\") in target(%s), please set only one tool for each tool kind.",
                toolkind, table.concat(table.wrap(program), "\", \""), target:fullname())
        end
        return true
    end, level = "error"}))
end
